/*	
 * Camera Picture Copy (campico)
 *
 * This will copy pictures from the mounted camera (i.e. usb mass-storage) directory,
 *  and copy the files to specified destination directory - more properly named.
 *
 * Copyright (c) 2002-2003 Torgeir Hansen <cpc@trenger.ro> of the fff (http://fff.sliter.nu)
 * Contributions (c) 2010 Kai Engert kaie at kuix dot de
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include "campico.h"

char campico_c_ver[] = "campico v 0.4.0-1";

//
// the mother of all func's ;)
//
int main(int argc, char **argv) {
	char ch;
	struct stat statbuf;
	
	// get user's wishes
	while((ch = getopt(argc, argv, "?hermnpvs:d:c:")) != -1) {
		switch (ch) {
			case 'c':
				cfgfile = strdup(optarg);
				break;
			case 'e':
				useexif = 1;
				break;
			case 's':
				sourcedir = strdup(optarg);
				break;
			case 'd':
				destdir = strdup(optarg);
				break;
			case 'r':
				recurse = 1;
				break;
			case 'm':
				movefiles = 1;
				break;
			case 'p':
				prunesource = 1;
				break;
			case 'n':
				dryrun = 1;
				break;
			case 'v':
				if (verbose == -1)
					verbose++;
				verbose++;
				break;
			case 'h':
			case '?':
				printhelp(argv[0]);
				exit(0);
			default:
				fprintf(stderr,"Oops, invalid option!\n");
				printhelp(argv[0]);
				exit(1);
		}
	}
	
	// try to read a config-file if it exists!
	//  I don't really care if it fails ;]
	readconf();
	
	// sanity for our variables; it's done this way to ease the whole configfile-thingie ;]
	if (useexif == -1)
		useexif = 0;
	if (verbose == -1)
		verbose = 0;
	if (recurse == -1)
		recurse = 0;
	if (movefiles == -1)
		movefiles = 0;
	if (prunesource == -1)
		prunesource = 0;
	if (dryrun == -1)
		dryrun = 0;
	
	// if user did not specify destination directory; use $HOME/pictures
	if (!destdir || !strlen(destdir)) {
		destdir = malloc(strlen(getenv("HOME")) + 20);
		sprintf(destdir,"%s/pictures",getenv("HOME"));
	}
	if ( *(destdir + strlen(destdir)-1)  == '/')
		*(destdir + strlen(destdir)-1) = '\0';
	
	if (!sourcedir || !strlen(sourcedir)) {
		sourcedir = strdup("/mnt/cam");
	}
	if ( *(sourcedir + strlen(sourcedir)-1)  == '/')
		*(sourcedir + strlen(sourcedir)-1) = '\0';
	
	
	//
	// check that we have proper access to source
	//
	if (stat(sourcedir,&statbuf)) {
		fprintf(stderr,"Error: cannot read source directory (%s): %s\n",sourcedir,strerror(errno));
		exit(1);
	}
	if (!S_ISDIR(statbuf.st_mode)) {
		fprintf(stderr,"Error: source (%s) is not a directory!\n", sourcedir);
		exit(1);
	}
	if (access(sourcedir,R_OK|X_OK)) {
		fprintf(stderr,"Error: cannot read source directory (%s): %s\n",sourcedir,strerror(errno));
		exit(1);
	}
	
	//
	// check that we have proper access to dest
	// 
	if (stat(destdir,&statbuf)) {
		fprintf(stderr,"Error: cannot read dest directory (%s): %s\n",destdir,strerror(errno));
		if (!recurse)
			fprintf(stderr,"Perhaps you forgot to create it?\n");
		exit(1);
	}
	if (!S_ISDIR(statbuf.st_mode)) {
		fprintf(stderr,"Error: dest (%s) is not a directory!\n",destdir);
		exit(1);
	}
	if (access(destdir,W_OK|X_OK)) {
		fprintf(stderr,"Error: cannot write dest directory (%s): %s\n",destdir,strerror(errno));
		exit(1);
	}
	
	// do the copying
	if (docopy(sourcedir,destdir)) {
		fprintf(stderr,"An error occured while copying files!\n");
		exit(1);
	}
	
	if (imagecount)
		printf("Done, copied %lu files (%.02f MB)\n",imagecount,  ((float) bytecount / (1024*1024)));
	else {
		printf("Nothing to do!\n");
		if (!recurse) 
			printf("Perhaps you need to specify -r?\n");
	}
	
	return 0;
} // end, main()

//
// help the user to know what to do
//
void printhelp(char *name) {
	printf("This is %s, version %s\n",PACKAGE,VERSION);
	printf("Usage: %s [OPTIONS]\n",name);
	printf("\t-h\t\t show this helpscreen.\n");
	printf("\t-v\t\t be verbose. Use twice for greater effect.\n");
	printf("\t-c CFGFILE\t read configuration from CFGFILE (default: ~/.cpcrc),\n");
	printf("\t\t\t  see campico(1) for example configuration.\n");
	printf("\t-s SOURCEDIR\t grab pictures from SOURCEDIR. (default: /mnt/cam)\n");
	printf("\t-d DESTDIR\t put files into DESTDIR. (default: ~/pictures)\n");
	printf("\t-r\t\t read sourcedir recursively.\n");
#ifdef HAVE_LIBEXIF
	printf("\t-e\t\t read exif tag from imagefiles. (default: off)\n");
#else
	printf("\t-e\t\t read exif tag from imagefiles. (Not available, need libexif)\n");
#endif
	printf("\t-m\t\t move files (if copying succeeds, delete source file).\n");
	printf("\t-p\t\t don't copy. prune source files, if destination file already exists.\n");
	printf("\t-n\t\t dry run (explain what would be done, but don't do it)\n");
	printf("\n");
	printf("Please send bug-reports and comments to kaie at kuix dot de\n");
	printf("\n");
	
	return;
} // end, printhelp()

int get_file_md5(const char *filename, char *output_md5_string)
{
  unsigned char md[MD5_DIGEST_LENGTH];
  int fd;
  MD5_CTX c;
  long bytesread;
  int r, n;
  const unsigned bufsize = 4096;
  char buf[bufsize];
  
  fd = open(filename,O_RDONLY);
  if (fd == -1)
    return -1;

  MD5_Init(&c);

  bytesread = 0;
  while ( (r = read(fd,buf,bufsize)) > 0) {
    bytesread += r;
    MD5_Update(&c, buf, r);
  }

  close(fd);
  MD5_Final(md, &c);

  output_md5_string[MD5_DIGEST_LENGTH] = 0;
  for (n = 0; n < MD5_DIGEST_LENGTH; n++)
    sprintf(output_md5_string+2*n, "%02x", md[n]);

  return 0;
}

//
// copy files from source; and write to dest/YYYYMMDD/HHMM-X.{extensionN}
//
int docopy (char *source, char *dest) {
	DIR *sourcedirfh = NULL, *destdirfh = NULL;
	struct dirent *dire = NULL;
	struct tm *timebuf;
	struct stat statbuf, filestat;
	struct utimbuf utimebuf;
	char filebuf[4096], *ext = NULL, *tmp = NULL, destdir[4096], 
	     tmpfile[8096], destfile[8096], datachunk[4096];
	int i,fh = -1, sourcefh, destlen = 0, need_mkdir;
	long bytesread;
	char dateprefix[21];
	char filetemplate[8001];
#ifdef HAVE_LIBEXIF
	ExifData *ed = NULL;
	ExifEntry *entry;
	char tempdate[21];
	int invalidtime = 0;
#endif
        char source_hash[1+MD5_DIGEST_LENGTH*2];
	
	if ( (sourcedirfh = opendir(source)) == NULL) {
		fprintf(stderr,"Error: could not open source directory: %s\n",strerror(errno));
		return 1;
	}
	if ( (destdirfh = opendir(dest)) == NULL) {
		fprintf(stderr,"Error: could not open dest directory: %s\n", strerror(errno));
		return 1;
	}
	
	destlen = strlen(dest) +1; // also count trailing '/'
	
	// read all files
	while ( (dire = readdir(sourcedirfh)) != NULL) {
		ext = NULL;
		if (dire->d_name[0] == '.') {
			if (dire->d_name[1] == '\0')
				continue;
			if (dire->d_name[1] == '.') {
				if (dire->d_name[2] == '\0')
					continue;
			}
		}
		snprintf(filebuf,4095,"%s/%s", source, dire->d_name);
		if (stat(filebuf,&filestat)) {
			fprintf(stderr,"Error: cannot read sourcefile (%s): %s\n",filebuf,strerror(errno));
			continue;
		}
		if (S_ISDIR(filestat.st_mode)) {
			if (recurse) {
				if (docopy(filebuf, dest))
					return 1;
				continue;
			} else {
				// silently ignore subdirectories
				continue;
			}
		}
		
		// get file-extension; if any. if there is no extension; use the actual filename.
		ext = dire->d_name;
		while ( (tmp = strstr(ext,".")) != NULL) {
			ext = tmp +1;
		}
		
		// get fileinfo
		timebuf = localtime(&filestat.st_mtime);
		
		// build directory and filename. (these will most likely be changed if we are using exif tags)
		snprintf(destdir,4095,"%s/%d-%.02d-%.02d",
				dest,timebuf->tm_year + 1900,timebuf->tm_mon+1,timebuf->tm_mday);
		snprintf(dateprefix,20,"%d-%.02d-%.02d-",
				timebuf->tm_year + 1900,timebuf->tm_mon+1,timebuf->tm_mday);

                snprintf(filetemplate,8000,"%.02d_%.02d_%.02d",timebuf->tm_hour,timebuf->tm_min,timebuf->tm_sec);
		
#ifdef HAVE_LIBEXIF
		
		if ( useexif && ((ed = exif_data_new_from_file (filebuf)) == NULL) ) {
			if (useexif)
				fprintf(stderr,"Warning: '%s' does not contain EXIF data, using file creation time!\n",filebuf);
		} else if (useexif) {
			// get exif information
			
			// find the proper entry
			for (i = 0; i < EXIF_IFD_COUNT; i++) {
				entry = exif_content_get_entry (ed->ifd[i], EXIF_TAG_DATE_TIME);
				if (entry) {
					break;
				}
			}
					
			if (entry) {
				memset((void*)tempdate,0,21);
				exif_entry_get_value(entry,tempdate,20);
				
				// check that it's a valid Ascii format, if not it's new to me;
				//  warn and tell them to email me.
				if (entry->format != 2) {
					fprintf(stderr,
						"WARNING: Found exif DateTime tag that is not in ASCII format.\n");
					fprintf(stderr, " Please report this to kaie at kuix dot de, ");
					fprintf(stderr, "along with this information:\n");
					fprintf(stderr," Format = %d\n",entry->format);
					fprintf(stderr," Value = %s\n",tempdate);
					fprintf(stderr,"Disable the usage of exif tags to complete your operation!\n");
					exit(1);
				}
				
				// get it's value
				if (!strcmp(tempdate,"0000:00:00 00:00:00")) {
					invalidtime = 1;
				}
				
				// insert year, verify that it's numeric!
				for (i = 0; i < 4; i++) {
					if ((tempdate[i] >= '0') && (tempdate[i] <= '9')) {
						destdir[destlen+i] = tempdate[i];
                                                dateprefix[i] = tempdate[i];
					} else
						invalidtime = 1;
				}
				
				// insert month
				if ((tempdate[5] >= '0') && (tempdate[5] <= '9')) {
					destdir[destlen + 5] = tempdate[5];
                                        dateprefix[5] = tempdate[5];
                                }
				else
					invalidtime = 1;
				if ((tempdate[6] >= '0') && (tempdate[6] <= '9')) {
					destdir[destlen + 6] = tempdate[6];
                                        dateprefix[6] = tempdate[6];
                                }
				else
					invalidtime = 1;
				
				// insert day
				if ((tempdate[8] >= '0') && (tempdate[8] <= '9')) {
					destdir[destlen + 8] = tempdate[8];
                                        dateprefix[8] = tempdate[8];
                                }
				else
					invalidtime = 1;
				if ((tempdate[9] >= '0') && (tempdate[9] <= '9')) {
					destdir[destlen + 9] = tempdate[9];
                                        dateprefix[9] = tempdate[9];
                                }
				else
					invalidtime = 1;
				
				
				//
				// fix filename
				//
				
				// set hour
				if ((tempdate[11] >= '0') && (tempdate[11] <= '9'))
					filetemplate[0] = tempdate[11];
				else
					invalidtime = 1;
				if ((tempdate[12] >= '0') && (tempdate[12] <= '9'))
					filetemplate[1] = tempdate[12];
				else
					invalidtime = 1;
				
				// set minute
				if ((tempdate[14] >= '0') && (tempdate[14] <= '9'))
					filetemplate[3] = tempdate[14];
				else
					invalidtime = 1;
				if ((tempdate[15] >= '0') && (tempdate[15] <= '9'))
					filetemplate[4] = tempdate[15];
				else
					invalidtime = 1;
				
				// set second
				if ((tempdate[17] >= '0') && (tempdate[17] <= '9'))
					filetemplate[6] = tempdate[17];
				else
					invalidtime = 1;
				if ((tempdate[18] >= '0') && (tempdate[18] <= '9'))
					filetemplate[7] = tempdate[18];
				else
					invalidtime = 1;
				
				// fallback
				if (invalidtime) {
					fprintf(stderr,"Invalid DateTime tag in file '%s' (%s)!\n",filebuf, tempdate);
					// since the input date is corrupt, use the original!
					snprintf(destdir,4095,"%s/%d-%.02d-%.02d",
						dest,timebuf->tm_year + 1900,timebuf->tm_mon+1,timebuf->tm_mday);
                                        snprintf(dateprefix,20,"%d-%.02d-%.02d-",
                                                timebuf->tm_year + 1900,timebuf->tm_mon+1,timebuf->tm_mday);
				}
				
			} else {
				snprintf(destdir,4095,"%s/%d-%.02d-%.02d",
					dest,timebuf->tm_year + 1900,timebuf->tm_mon+1,timebuf->tm_mday);
                                snprintf(dateprefix,20,"%d-%.02d-%.02d-",
                                        timebuf->tm_year + 1900,timebuf->tm_mon+1,timebuf->tm_mday);
                        }
			
			exif_data_unref(ed);
		}	
#endif	

		need_mkdir = 0;

	      if (!prunesource) {
		// check if we have to create the destination directory.
		if (stat(destdir,&statbuf)) {
			if (errno == ENOENT) {
			  need_mkdir = 1;
			  if (!dryrun) {
				if (verbose)
					printf("New directory: %s\n",destdir);
				// create!
				if (mkdir(destdir,0700)) {
					fprintf(stderr,"Failed to create destination directory (%s): %s\n",
							destdir,strerror(errno));
					return 1;
				}
			  }
			} else {
				// bork, bork!
				fprintf(stderr,"Failed to create destination directory (%s): %s\n",
						destdir,strerror(errno));
				return 1;
			}
		}
	      }


                
                if (get_file_md5(filebuf, source_hash)) {
                  fprintf(stderr, "failed to hash source file (%s): %s\n", filebuf, strerror(errno));
                  return 1;
                }

		
                      snprintf(tmpfile,8095,"%s/%s%s-%s.%s.tmp",destdir,dateprefix,filetemplate,source_hash,ext);
                      snprintf(destfile,8095,"%s/%s%s-%s.%s",destdir,dateprefix,filetemplate,source_hash,ext);
		      
	    if (stat(destfile,&statbuf) == 0) {
	      if (prunesource) {
		if (dryrun) {
		      printf("Would delete %s, destination %s already exists\n",filebuf,destfile);
		}
		else {
		  if (unlink(filebuf)) {
		    fprintf(stderr,"Failed to delete source file (%s): %s\n",filebuf,strerror(errno));
		  }
		  else {
		    if (verbose > 1)
		      printf("Deleted %s, destination %s already exists\n",filebuf,destfile);
		  }
		}
	      }
	      else {
		if (verbose > 1) {
		  printf("Skipping %s, destination %s already exists\n",filebuf,destfile);
		}
	      }
	      continue;
	    }
	    
	    if (prunesource)
	      continue;
                      
	    if (!dryrun) {
		// try to create the tmp file used while copying, overwrite left over tmp file
		fh = open(tmpfile,O_CREAT|O_TRUNC|O_WRONLY,0600);
		if (fh < 0) {
			  fprintf(stderr,"Could not create tmp file (%s): %s\n",tmpfile, strerror(errno));
			  return 1;
		}
		
		// open the sourcefile
		if ( (sourcefh = open(filebuf,O_RDONLY)) == -1) {
			fprintf(stderr,"Could not open source file (%s): %s\n",filebuf,strerror(errno));
			return 1;
		}
		
		// copy to dest
		// XXX: if I get around to check for linux-sendfile; this can be reactivated with some
		// ifdef's to include it when it should be :]
		//if (sendfile(fh,sourcefh,NULL, filestat.st_size) != filestat.st_size) {
		//	fprintf(stderr,"Could not write dest file (%s): %s\n",tmpfile,strerror(errno));
		//	return 1;
		//}
		bytesread = 0;
		while ( (i = read(sourcefh,datachunk,4096)) > 0) {
			bytesread += i;
			if (write(fh,datachunk,i) != i) {
				fprintf(stderr,"Error while writing picture (%s): %s\n",tmpfile,strerror(errno));
				close(sourcefh);
				close(fh);
				unlink(tmpfile);
				return 1;
			}
		}
		
		// close files
		close(sourcefh);
		close(fh);
		
		// fix modification time for file
		utimebuf.actime = filestat.st_atime;
		utimebuf.modtime = filestat.st_mtime;
		if (utime(tmpfile,&utimebuf)) {
			fprintf(stderr,"Failed to set modification date/time for file (%s): %s\n",tmpfile,strerror(errno));
			unlink(tmpfile);
			return 1;
		}
		
		if (rename(tmpfile,destfile)) {
			fprintf(stderr,"Failed to move tmp file to final name (%s): %s\n",destfile,strerror(errno));
			unlink(tmpfile);
			return 1;
		}

		if (movefiles)
		if (unlink(filebuf)) {
			fprintf(stderr,"Failed to delete source file (%s): %s\n",filebuf,strerror(errno));
		}
	    }
		
	    // only show file copy if we are very verbose
	    if (verbose > 1) {
	      printf("%s %s to %s%s\n", (dryrun?"Would copy":"Copied"),
		      filebuf,destfile, (need_mkdir && dryrun?" (new directory)":""));
	      if (movefiles)
		printf("%s %s\n", (dryrun?"Would delete":"Deleted"),
			filebuf);
	    }
	    
	    // update counters
	    imagecount++;
	    bytecount += filestat.st_size;
		
	}
	closedir(sourcedirfh);
	closedir(destdirfh);
	
	return 0;
} // end, docopy()

//
// attempt to read configuration, if present.
//
int readconf(void) {
	FILE *fh = NULL;
	char *configfile = NULL;
	char inbuf[512], *tmp = NULL, *tmp2 = NULL, ch;
	char *t1,*t2,*t3;
	
	// sanity
	if (!cfgfile) {
		cfgfile = strdup(".cpcrc");
	}
	
	configfile = malloc(strlen(getenv("HOME")) + strlen(cfgfile) + 2);
	sprintf(configfile,"%s/%s",getenv("HOME"), cfgfile);
	
	fh = fopen(configfile,"r");
	if (!fh) {
		if (errno != ENOENT) {
			fprintf(stderr,"Warning: unable to open configuration file (%s): %s\n",configfile,strerror(errno));
			free(configfile);
			return 1;
		}
		// return quiet if there is no configuration-file
		return 0;
	}
	while (!feof(fh)) {
		memset((char *)inbuf,0,512);
		tmp = inbuf;
		if (fgets(inbuf,511,fh) <= 0)
			break;
		
		// ignore leading spaces/tabs
		while ((*tmp == ' ') || (*tmp == '\t'))
			tmp++;
		
		// if file starts with a hash; ignore this line!
		if (*tmp == '#')
			continue;
		
		// ensure that we check the entire word, not just partial if long words
		t1 = strstr(tmp," ");
		t2 = strstr(tmp,"\t");
		t3 = strstr(tmp,"=");
		ch = '\0';
		tmp2 = NULL;
		if (t1) {
			tmp2 = t1;
			if (t2 && t2 < tmp2)
				tmp2 = t2;
			if (t3 && t3 < tmp2)
				tmp2 = t3;
		} else if (t2) {
			tmp2 = t2;
			if (t1 && t1 < tmp2)
				tmp2 = t1;
			if (t3 && t3 < tmp2)
				tmp2 = t3;
		} else if (t3) {
			tmp2 = t3;
			if (t1 && t1 < tmp2)
				tmp2 = t1;
			if (t2 && t2 < tmp2)
				tmp2 = t2;
		}
		
		if (tmp2) {
			ch = *tmp2;
			*tmp2 = '\0';
		}
		
		// default source?
		if (!strcasecmp(tmp,"sourcedir")) {
			if (!sourcedir) {
				if (tmp2) { *tmp2 = ch; }
				sourcedir = getparam(tmp, "sourcedir");
				if (!sourcedir)
					continue;
				if (verbose != -1) {
					fprintf(stderr,"Info: source directory set to %s\n",sourcedir);
				}
			}
			continue;
		}
		
		// default dest?
		if (!strcasecmp(tmp,"destdir")) {
			if (!destdir) {
				if (tmp2) { *tmp2 = ch; }
				destdir = getparam(tmp, "destdir");
				if (!destdir)
					continue;
				if (verbose != -1) {
					fprintf(stderr,"Info: destination directory set to %s\n",destdir);
				}
			}
			continue;
		}
		
		// default recurse?
		if (!strcasecmp(tmp,"recurse")) {
			if (recurse == -1) {
				if (tmp2) { *tmp2 = ch; }
				tmp2 = getparam(tmp, "recurse");
				if (tmp2) {
					if (!strcasecmp(tmp2,"yes"))
						recurse = 1;
					if (!strcasecmp(tmp2,"1"))
						recurse = 1;
					
					free(tmp2);
				}
			}
			continue;
		}
		
		// default be verbose?
		if (!strcasecmp(tmp,"verbose")) {
			if (verbose == -1) {
				if (tmp2) { *tmp2 = ch; }
				tmp2 = getparam(tmp, "verbose");
				if (tmp2) {
					verbose = atoi(tmp2);
					if (!strcasecmp(tmp2,"yes"))
						verbose = 1;
					
					free(tmp2);
				}
			}
			continue;
		}
		
		// default useexif?
		if (!strcasecmp(tmp,"readexif")) {
			if (useexif== -1) {
				if (tmp2) { *tmp2 = ch; }
				tmp2 = getparam(tmp, "readexif");
				if (tmp2) {
					if (!strcasecmp(tmp2,"1"))
						useexif = 1;
					if (!strcasecmp(tmp2,"yes"))
						useexif = 1;
					
					free(tmp2);
				}
			}
			continue;
		}
		
		// default movefiles?
		if (!strcasecmp(tmp,"movefiles")) {
			if (movefiles== -1) {
				if (tmp2) { *tmp2 = ch; }
				tmp2 = getparam(tmp, "movefiles");
				if (tmp2) {
					if (!strcasecmp(tmp2,"1"))
						movefiles = 1;
					if (!strcasecmp(tmp2,"yes"))
						movefiles = 1;
					
					free(tmp2);
				}
			}
			continue;
		}
	}
	fclose(fh);
	
	return 0;
} // end, readconf()

//
// extract parameter value for given keyword
//
char *getparam(char *line, char *keyword) {
	char *tmp = NULL, *tmp2 = NULL;
	
	// check for a '='
	tmp = strstr(line,"=");
	if (!tmp)
		return NULL;
	
	// ignore the '='
	tmp++;
	
	// ignore leading spaces/tabs
	while ((*tmp == ' ') || (*tmp == '\t'))
		tmp++;
	
	// check if it's encapsulated within "'s
	tmp2 = tmp;
	if (*tmp2 == '"') {
		// skip the '"'
		tmp2++;
		tmp++;
		
		while ((*tmp2 != '#') && (*tmp2 != '\0') && (*tmp2 != '"'))
			tmp2++;
		
		if (*tmp2 == '"') {
			*tmp2 = '\0';
		} else {
			// this appears to be a invalid line; i.e: source = "blah blah #\n
			if (verbose != -1) {
				fprintf(stderr,"Warning: invalid '%s =' in configuration!\n", keyword);
				return NULL;
			}
		}
	} else {
		// terminate at first ' ', '\t' or '#' 
		while ((*tmp2 != '\0') && (*tmp2 != '#') && (*tmp2 != ' ') && (*tmp2 != '\t'))
			tmp2++;
		if (*tmp2 != '\0') {
			*tmp2 = '\0';
		}
	}
	
	return strdup(tmp);
} // end, getparam()

